// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *				Hardware
 *
 * The present file describes blocks of hardware (units) the VM_UNT
 * is made of, namely, CPU, diagnostic panel, memory, channels, etc.
 * See the document vm_unt.txt for details
 *
 ************************************************************************
 */

#ifndef _hardware_h
#define _hardware_h 1
#pragma interface

#include "mult_proc.h"
#include "instruction_set.h"

/*
 *------------------------------------------------------------------------
 *		   	Printing diagnostic info
 */

class DiagnosticPanel
{
				// It is necessary for output from one process
				// not to get messed up with that of 
				// other processes
  static Semaphore printing_diagnostics;

public:
  short is_tracing;

  DiagnosticPanel(void);
  virtual ~DiagnosticPanel(void) {}

				// This routine displays the message
				// only if tracing is on
  void single_message(const char * message,...);	
				// This program always displays the
				// message
  void console(const char * message,...);
				// The following function bracket the
				// block of diagnostic info that
				// should not be broken by other processes
				// output
  void begin_printing(void);
  void end_printing(void);
};

/*
 *------------------------------------------------------------------------
 *			Virtual and Real memory
 */

typedef short Word;
typedef unsigned short Address;


				// This is a physical memory of the computer
class Memory : public DiagnosticPanel
{
  Word array[1024];			// Memory itself
  int drum_fh;				// File handle of a drum device

public:
  const Address hiaddr = 1023;		// The largest memory address
  const short pagesize = 32;		// Size of the memory page in words
					// = size of the drum sector
  const short no_drum_sectors = 256;
  const short no_pages;			// No of physical pages into the memory

public:
  Memory(void);
  ~Memory(void)	{}

					// Basic operation to get hold of a 
					// word of memory given its address
  struct MemoryAnswer {char out_of_range; Word& contents;};
  MemoryAnswer operator [] (const Address addr)
    { if( addr > hiaddr ) 
      { static Word dummy; MemoryAnswer ans = {1,dummy}; return ans;}
      else {MemoryAnswer ans = {0,array[addr]}; return ans;} }


  void dump(const Address from, const Address to); // Dump the memory context

  void drum_open(const char * file_name);
  void drum_read(const int sector_addr, const int pageframeno);
  void drum_write(const int pageframeno,const int sector_addr);
				// Load a program from the drum into the
				// memory and return its starting address
  Address drum_load(const int pageframeno);
};


struct MMUContext		// This is the context of the MMU
{
  Address page_table_addr;		// Page table addr in the phys memory
  short	  page_table_len;		// No of entries in the page table
					// (size of the virtual space in pages)
  char enabled_address_translation;

  struct VirtualMemoryAddr {
    unsigned short
      offset  :	 5,			// Offset within the page
      page_no :  5,			// Page no
      reserved:  6;
  };

  struct PageTableEntry {
    unsigned short
      phys_page_no: 5,			// No for the physical page
      reserved:	    3,			// Reserved (always 0)
      valid:	    1,			// 0 means no access is possible
      any_access:   1,			// 1 means READ/WRITE/EXEC is allowed
      read_only:    1,			// 1 means WRITE access is forbidden
      exec_only:    1,			// 1 means only EXEC access is allowed
      was_ref:	    1,			// The page was referenced
      was_written:  1;			// The page was modified by WRITE
  };


  MMUContext(void);
  virtual ~MMUContext(void) {}
};

				// This part is responsible for the virtual-to
				// real address translation
class MemoryManagementUnit : public MMUContext
{
public:
  Memory& memory;			// Memory this unit manages

public:
  enum MemoryFault {None, OutofRange, WrongPageTable, WrongVPageNo,
		    PageInvalidated, NoAccess, ReadOnly, ExecOnly};
private:
  MemoryFault mem_error;
  VirtualMemoryAddr current_VA;			// Address under translation

public:
  MemoryManagementUnit(Memory& phys_memory);
  virtual ~MemoryManagementUnit(void) {}

  Word fetch(const Address addr);	// Fetch a word from memory in
  					// an EXECUTE memory mode
  Word get(const Address addr);		// Read a word from memory
					// Write a word to memory
  void put(const Address addr,const Word data);

  MemoryFault what_fault(void) const	{ return mem_error; }
  int got_fault(void) const		{ return mem_error != None; }
  VirtualMemoryAddr q_current_VA(void)	{ return current_VA; } 
  void clear_error(void)		{ mem_error = None; }

					// Dump the physical memory
  void dump_phys_memory(const Address from, const Address to)
  	{ memory.dump(from,to); }
				// Load a program from the drum into the
				// memory and return its starting address
				// The program is in absolute format
  Address absolute_drum_load(const int pageframeno)
  	{ return memory.drum_load(pageframeno); }

  void dump(void) const;		// Dump the info in the table
};

/*
 *------------------------------------------------------------------------
 *			A Central Processing Unit
 */

struct Context
{
  Word Registers[16];			// Registers pool
  Word pc;				// Program Counter

  Context(void);			// Constructor (initializer)
  ~Context(void) {}

  Word& operator [] (const int reg_number);	// Get a ref to the register
  void copy_context(const Context& a_context);
  void dump(void) const;			// Dump the context
};

class CentralProcessingUnit : 
	public Context, public MemoryManagementUnit, public DiagnosticPanel
{
  Semaphore stopped;			// 0 if running, 1 can run, -1 stopped

				// Instruction being executed
  enum OpCodes opcode;			// Opcode of the current instruction
  short inst_length;			// Length in words
  InstructionWord1 iword1;
  InstructionWord2 iword2;
  short svcop;				// The code of the SVC instruction
  Address EA;				// Effective address

  int execute_instruction(void);	// Execute the current instruction
					// Return true if can continue

public:
  short clock;				// Generate a trap when going negative

  enum TRAPCODE { NONE, ILLEGOP, SVCCALL, CLOCKINT, MEMORY };
private:
  Semaphore trap_signal;		// V(trap_signal) when any traps occurs
  enum TRAPCODE trap_code;

public:
  CentralProcessingUnit(Memory & cpu_memory);
  ~CentralProcessingUnit(void) {}

  void start(void);			// Start the CPU from the current value
					// of PC. Do nothing if already running
  void stop(void);			// Stop the CPU after executing the
					// current instruction

				// CPU status inquires
  int is_running(void)			{ return !stopped.awaited(); }
  void wait_for_trap(void)		{ trap_signal--; }

  TRAPCODE q_trap_code(void) const	{ return trap_code; }
  OpCodes  q_opcode(void)    const	{ return opcode;    }
  short    q_svcop(void)     const	{ return svcop;     }
  Word     q_RegA(void)      const	{ return iword1.areg; }
  Address  q_EA(void)        const	{ return EA; }

  void dump(void) const;		// Tell what's going on
};

/*
 *------------------------------------------------------------------------
 *			An I/O channel processing unit
 */

class IOChannel;

				// I/O bus all the channels are hooked to
class IOBus : public DiagnosticPanel
{
					// Note, the numeration of channels
  const int max_no_channels = 4;	// starts from one
  unsigned short no_channels;
  IOChannel * channels[max_no_channels];

public:
  IOBus(void);
  ~IOBus(void) {}

					// Connects a channel to the i/o bus
					// and returns its number on the bus
  int connect_channel(IOChannel * channel);

					// Start a channel program on a 
					// specified channel. Return FALSE
					// if the operation was rejected 
					// upfront
  int start(const unsigned short channel_no, const Address ccw_ptr);

  void dump(void);			// Dump info about all the channels
};


class IOChannel : public DiagnosticPanel
{
private:
  Semaphore stopped;			// 0 if running, 1 can run, -1 stopped
  short channel_no;			// Starts with 1

public:
  Memory& memory;			// Memory to read/write to/from

private:
				// CCW being executed
  Address cpc;				// Channel program counter
  CCW  current_CCW;
  enum CCWOpCodes opcode;		// Opcode of the current CCW
  unsigned short buf_page;		// Page no of the I/O buffer

protected:
				// Current I/O status
  const int rec_size = Memory::pagesize;// For any I/O operation
  unsigned short rec_count;		// The number of the current records
					// to transfer
  Word io_buffer[rec_size];		// I/O buffer

  int file_handle;			// File handle of the UNIX device that
					// emulates the channel

private:
				// Transfer between the memory and the 
				// i/o buffer. The routines below return
				// FALSE if a problem occurred.
  int get_record(const int page_no);	// Get a record from memory
  int put_record(const int page_no);	// Put a record from memory

public:
  enum STATCODE { OK, CCWADDR, MEMORY, ILLEGOP, END_FILE, SEEKADDR, DEVICE };
private:
  enum STATCODE status;

  Semaphore intr_signal;		// V(intr_signal) when any channel
					// interrupt occurs
  STATCODE execute_CCW(void);		// Execute the current CCW and report
					// the status
protected:
				// This routine performs device-specific
				// operations. The routine should transfer
				// the io_buffer to/from the device and set
				// the rec_counter.
  enum DEVICE_OP {DEV_READ, DEV_SEEK, DEV_WRITE};
  virtual STATCODE device_driver(const DEVICE_OP dev_op) = 0;

public:
  IOChannel(Memory& main_memory, IOBus& io_bus);
  ~IOChannel(void) {}

  void start(Address channel_program_ptr);// Start the channel

				// Channel status inquires
  int q_channel_no(void) const		{ return channel_no; }
  int is_running(void)			{ return !stopped.awaited(); }
  void wait_for_interrupt(void)		{ intr_signal--; }

  STATCODE q_status(void) const		{ return status; }
  int q_buf_page(void) const		{ return buf_page; }

  void dump(void) const;		// Tell what's going on

				// Open the UNIX file the channel is associated
				// with
  void associate_with_unix_file(const char * file_name);
};

				// Device specific stuff
class CardReader : public IOChannel
{
  void * in_file;			// FILE * in_file

  virtual STATCODE device_driver(const DEVICE_OP dev_op);

public:
  CardReader(Memory& main_memory, IOBus& io_bus);
  ~CardReader(void) {}
};

class LinePrinter : public IOChannel
{
  virtual STATCODE device_driver(const DEVICE_OP dev_op);

public:
  LinePrinter(Memory& main_memory, IOBus& io_bus) : 
  		IOChannel(main_memory,io_bus) {}
  ~LinePrinter(void) {}
};

class HardDisk : public IOChannel
{
  const int hi_disk_addr = 1023;
  const int disk_sector_size = Memory::pagesize * sizeof(Word);  // in bytes

  virtual STATCODE device_driver(const DEVICE_OP dev_op);

public:
  HardDisk(Memory& main_memory, IOBus& io_bus) :
  		IOChannel(main_memory,io_bus) {}
  ~HardDisk(void) {}
};

#endif

