// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 * 		         Operations on semaphores
 *
 ************************************************************************
 */

#include "semaphores.h"
#pragma implementation

#include "myenv.h"
#include <std.h>

/*
 *------------------------------------------------------------------------
 *			Constructing the Semaphore
 */

Sema::Sema(void) : lock("Sema lock",1)
{
  owner = NIL_pid;
  value = 0;
}

/*
 *------------------------------------------------------------------------
 *			Elementary Semaphore operations
 */
				// Perform P-operation. If value goes
				// negative, return FALSE and queue the
				// pcb into the internal queue
int Sema::p(PCB& pcb)
{
  lock--;				// Start a critical section
  if( --value < 0 )
    waiting_list.append(pcb);
  register int result = (value >= 0);
  lock++;				// End the critical section
  return result;
}

				// Perform V-operation. If value was negative
				// get a process from waiting list and
				// return its PID. Otherwise return NIL_pid
PID Sema::v(void)
{
  PID pid_to_awake = NIL_pid;
  lock--;				// Start a critical section
  if( value++ < 0 )
    pid_to_awake = (waiting_list.get_from_head())->id;
  lock++;				// End the critical section
  return pid_to_awake;
}

				// Dispose of an active semaphore
void Sema::dispose(void)
{
  lock--;				// Start a critical section
  owner = NIL_pid;
  for(; value < 0; value++)
    (void)waiting_list.get_from_head();
  value = 0;
  lock++;				// End the critical section
}

				// Purge a process pid from a waiting queue,
				// (if any)
void Sema::purge(const PID pid)
{
  if( value >= 0 )
    return;				// No waiting queue for this semaphore

  lock--;				// Start a critical section
  if( waiting_list.purge_id(pid) )
    ++value;				// Increment the counter, if purged
  lock++;				// End the critical section
}

/*
 *------------------------------------------------------------------------
 *		     Constructing the semaphore table
 */

SemaphoreTable::SemaphoreTable(const int _nslots)
	: nslots(_nslots)
{
  assure(nslots > 4,"No of slots in a process table must be at least 5");
  slots = new Sema[nslots];

  register SID i;
  for(i=1; i<=nslots; i++)
  {
    Sema& sem = (*this)[i];
    sem.id = i;
    sem.key = 0;
    freeSEMs.append(sem);
  }
}


/*
 *------------------------------------------------------------------------
 *		     Elementary SemaphoreTable operations
 */

				// Get a Semaphore given its index
Sema& SemaphoreTable::operator [] (const SID id)
{
  assure(id != NIL_sid,"NIL index");
  if( id > nslots )
    _error("FATAL: Semaphore id %d is too big",id);
  return slots[id-1];
}

				// Check to see if SID is active
int SemaphoreTable::is_active(const SID id)
{
  if( id == NIL_sid || id > nslots || (*this)[id].owner == NIL_pid )
    return 0;
  return 1;
}

				// Get a new SID or NIL if it is not
				// possible
SID SemaphoreTable::new_semaphore(const int in_value, const PID owner)
{
  if( freeSEMs.is_empty() )
    return NIL_sid;
  Sema& sem = (*this)[freeSEMs.get_from_head()->id];
  sem.owner = owner;
  sem.value = in_value;
  return sem.id;
}

				// Print info about all active semaphores
void SemaphoreTable::dump(void)
{
  if( freeSEMs.q_no_elems() == nslots )
  {
    message("\nNo active semaphores\n");
    return;
  }

  message("\nActive semaphores:\n");
  register SID i;
  for(i=1; i<=nslots; i++)
  {
    Sema& sem = (*this)[i];
    if( sem.owner != NIL_pid )
      message("\tID = %d, process-owner = %d, value = %d\n",
	      i,sem.owner,sem.value);
  }
}

				// Get SID of a semaphore owned by a
				// given owner
SID SemaphoreTable::owned_by(const PID owner)
{
  register SID i;
  for(i=1; i<=nslots; i++)
  {
    Sema& sem = (*this)[i];
    if( sem.owner == owner )
      return sem.id;
  }
  return NIL_sid;
}

				// Dispose of a given semaphore
void SemaphoreTable::dispose(const SID sid)
{
  Sema& sem = (*this)[sid];
  assert( sem.owner != NIL_pid && sem.id == sid );
  sem.dispose();
  freeSEMs.append(sem);
}

				// Purge the process PID from the semaphore
				// queue
void SemaphoreTable::purge(const PID pid)
{
  register SID i;			// Purge from any active
  for(i=1; i<=nslots; i++)		// semaphore queue
  {
    Sema& sem = (*this)[i];
    if( sem.owner != NIL_pid )
      sem.purge(pid);
  }
}
